import { LamportClock as Clock } from "./clock.js";
import { Blocks } from "@dao-xyz/peerbit-block";
import { MaybeEncrypted, PublicSignKey, X25519PublicKey, PublicKeyEncryptionResolver, SignatureWithKey, Ed25519PublicKey } from "@dao-xyz/peerbit-crypto";
import { Encoding } from "./encoding.js";
import { Identity } from "./identity.js";
import { StringArray } from "./types.js";
export declare type MaybeEncryptionPublicKey = X25519PublicKey | X25519PublicKey[] | Ed25519PublicKey | Ed25519PublicKey[] | undefined;
export declare type EncryptionTemplateMaybeEncrypted = EntryEncryptionTemplate<MaybeEncryptionPublicKey, MaybeEncryptionPublicKey, MaybeEncryptionPublicKey | {
    [key: string]: MaybeEncryptionPublicKey;
}, // signature either all signature encrypted by same key, or each individually
MaybeEncryptionPublicKey>;
export interface EntryEncryption {
    reciever: EncryptionTemplateMaybeEncrypted;
    options: PublicKeyEncryptionResolver;
}
export declare function toBufferLE(num: bigint, width: number): Uint8Array;
export declare function toBigIntLE(buf: Uint8Array): bigint;
export declare type CanAppend<T> = (canAppend: Entry<T>) => Promise<boolean> | boolean;
export declare class Payload<T> {
    data: Uint8Array;
    _value?: T;
    constructor(props?: {
        data: Uint8Array;
        value?: T;
    });
    equals(other: Payload<T>): boolean;
    getValue(encoding?: Encoding<T>): T;
}
export interface EntryEncryptionTemplate<A, B, C, D> {
    metadata: A;
    payload: B;
    signatures: C;
    next: D;
}
export declare class Metadata {
    gid: string;
    clock: Clock;
    maxChainLength: bigint;
    constructor(properties?: {
        gid: string;
        clock: Clock;
        maxChainLength: bigint;
    });
}
export declare class Signatures {
    signatures: MaybeEncrypted<SignatureWithKey>[];
    constructor(properties?: {
        signatures: MaybeEncrypted<SignatureWithKey>[];
    });
    equals(other: Signatures): boolean;
}
export declare class Entry<T> implements EntryEncryptionTemplate<Metadata, Payload<T>, SignatureWithKey[], Array<string>> {
    _metadata: MaybeEncrypted<Metadata>;
    _payload: MaybeEncrypted<Payload<T>>;
    _next: MaybeEncrypted<StringArray>;
    _fork: MaybeEncrypted<StringArray>;
    _reserved: number[];
    _signatures?: Signatures;
    hash: string;
    _encryption?: PublicKeyEncryptionResolver;
    _encoding?: Encoding<T>;
    constructor(obj?: {
        payload: MaybeEncrypted<Payload<T>>;
        signatures?: Signatures;
        metadata: MaybeEncrypted<Metadata>;
        next: MaybeEncrypted<StringArray>;
        fork?: MaybeEncrypted<StringArray>;
        reserved?: number[];
        hash?: string;
    });
    init(props: {
        encryption?: PublicKeyEncryptionResolver;
        encoding: Encoding<T>;
    } | Entry<T>): Entry<T>;
    get encoding(): Encoding<T>;
    get metadata(): Metadata;
    getMetadata(): Promise<Metadata>;
    get gid(): string;
    getGid(): Promise<string>;
    getClock(): Promise<Clock>;
    get maxChainLength(): bigint;
    getMaxChainLength(): Promise<bigint>;
    get payload(): Payload<T>;
    getPayload(): Promise<Payload<T>>;
    getPayloadValue(): Promise<T>;
    get publicKeys(): PublicSignKey[];
    getPublicKeys(): Promise<PublicSignKey[]>;
    get next(): string[];
    getNext(): Promise<string[]>;
    /**
     * Will only return signatures I can decrypt
     * @returns signatures
     */
    get signatures(): SignatureWithKey[];
    /**
     * Will only return signatures I can decrypt
     * @returns signatures
     */
    getSignatures(): Promise<SignatureWithKey[]>;
    /**
     * Will only verify signatures I can decrypt
     * @returns true if all are verified
     */
    verifySignatures(): Promise<boolean>;
    static toSignable(entry: Entry<any>): Uint8Array;
    toSignable(): Uint8Array;
    equals(other: Entry<T>): boolean;
    delete(store: Blocks): Promise<void>;
    static createGid(seed?: string): Promise<string>;
    static create<T>(properties: {
        store: Blocks;
        gid?: string;
        gidSeed?: string;
        data: T;
        encoding?: Encoding<T>;
        canAppend?: CanAppend<T>;
        next?: Entry<T>[];
        clock?: Clock;
        pin?: boolean;
        encryption?: EntryEncryption;
        identity: Identity;
        signers?: ((data: Uint8Array) => Promise<SignatureWithKey>)[];
    }): Promise<Entry<T>>;
    /**
     * Get the multihash of an Entry.
     * @param ipfs An IPFS instance
     * @param entry Entry to get a multihash for
     * @returns {Promise<string>}
     * @example
     * const multfihash = await Entry.toMultihash(ipfs, entry)
     * console.log(multihash)
     * // "Qm...Foo"
     */
    static toMultihash<T>(store: Blocks, entry: Entry<T>, pin?: boolean): Promise<string>;
    /**
     * Create an Entry from a hash.
     * @param {IPFS} ipfs An IPFS instance
     * @param {string} hash The hash to create an Entry from
     * @returns {Promise<Entry<T>>}
     * @example
     * const entry = await Entry.fromMultihash(ipfs, "zd...Foo")
     * console.log(entry)
     * // { hash: "Zd...Foo", payload: "hello", next: [] }
     */
    static fromMultihash<T>(store: Blocks, hash: string): Promise<Entry<any>>;
    /**
     * Compares two entries.
     * @param {Entry} a
     * @param {Entry} b
     * @returns {number} 1 if a is greater, -1 is b is greater
     */
    static compare<T>(a: Entry<T>, b: Entry<T>): number;
    /**
     * Check if an entry equals another entry.
     * @param {Entry} a
     * @param {Entry} b
     * @returns {boolean}
     */
    static isEqual<T>(a: Entry<T>, b: Entry<T>): boolean;
    /**
     * Check if an entry is a parent to another entry.
     * @param {Entry} entry1 Entry to check
     * @param {Entry} entry2 The parent Entry
     * @returns {boolean}
     */
    static isDirectParent<T>(entry1: Entry<T>, entry2: Entry<T>): boolean;
    /**
     * Find entry's children from an Array of entries.
     * Returns entry's children as an Array up to the last know child.
     * @param {Entry} entry Entry for which to find the parents
     * @param {Array<Entry<T>>} values Entries to search parents from
     * @returns {Array<Entry<T>>}
     */
    static findDirectChildren<T>(entry: Entry<T>, values: Entry<T>[]): Entry<T>[];
}
