import { EntryIndex } from "./entry-index.js";
import { GSet } from "./g-set.js";
import * as Sorting from "./log-sorting.js";
import { EntryFetchAllOptions, EntryFetchOptions } from "./entry-io.js";
import { EncryptionTemplateMaybeEncrypted, Entry, Payload, CanAppend } from "./entry.js";
import { HLC, LamportClock, Timestamp } from "./clock.js";
import { PublicKeyEncryptionResolver, SignatureWithKey } from "@dao-xyz/peerbit-crypto";
import { Encoding } from "./encoding.js";
import { Identity } from "./identity.js";
import { HeadsIndex } from "./heads.js";
import { Blocks } from "@dao-xyz/peerbit-block";
export interface PruneOptions {
    maxLength: number;
    cutToLength: number;
}
export declare type LogOptions<T> = {
    encryption?: PublicKeyEncryptionResolver;
    encoding?: Encoding<T>;
    logId?: string;
    entries?: Entry<T>[];
    heads?: any;
    clock?: LamportClock;
    sortFn?: Sorting.ISortFunction;
    concurrency?: number;
    prune?: PruneOptions;
};
/**
 * @description
 * Log implements a G-Set CRDT and adds ordering.
 *
 * From:
 * "A comprehensive study of Convergent and Commutative Replicated Data Types"
 * https://hal.inria.fr/inria-00555588
 */
export declare class Log<T> extends GSet {
    _sortFn: Sorting.ISortFunction;
    _storage: Blocks;
    _id: string;
    _identity: Identity;
    _entryIndex: EntryIndex<T>;
    _headsIndex: HeadsIndex<T>;
    _nextsIndex: {
        [key: string]: Set<string>;
    };
    _nextsIndexToHead: {
        [key: string]: Set<string>;
    };
    _length: number;
    _prune?: PruneOptions;
    _encryption?: PublicKeyEncryptionResolver;
    _encoding: Encoding<T>;
    _hlc: HLC;
    joinConcurrency: number;
    constructor(store: Blocks, identity: Identity, options?: LogOptions<T>);
    /**
     * Returns the length of the log.
     * @return {number} Length
     */
    get length(): number;
    /**
     * Returns the values in the log.
     * @returns {Array<Entry<T>>}
     */
    get values(): Entry<T>[];
    /**
     * Returns an array of heads.
     * @returns {Array<Entry<T>>}
     */
    get headsIndex(): HeadsIndex<T>;
    /**
     * Don use this anywhere performance matters
     */
    get heads(): Entry<T>[];
    /**
     * Returns an array of Entry objects that reference entries which
     * are not in the log currently.
     * @returns {Array<Entry<T>>}
     */
    get tails(): Entry<T>[];
    /**
     * Returns an array of hashes that are referenced by entries which
     * are not in the log currently.
     * @returns {Array<string>} Array of hashes
     */
    get tailHashes(): string[];
    /**
     * Get local HLC
     */
    get hlc(): HLC;
    /**
     * Set the identity for the log
     * @param {Identity} [identity] The identity to be set
     */
    setIdentity(identity: Identity): void;
    /**
     * Find an entry.
     * @param {string} [hash] The hashes of the entry
     * @returns {Entry|undefined}
     */
    get(hash: string): Entry<T> | undefined;
    /**
     * Checks if a entry is part of the log
     * @param {string} hash The hash of the entry
     * @returns {boolean}
     */
    has(entry: Entry<string> | string): boolean;
    traverse(rootEntries: Entry<T>[], amount?: number, endHash?: string): {
        [key: string]: Entry<T>;
    };
    getPow2Refs(heads: Entry<T>[], pointerCount?: number): Entry<T>[];
    getHeadsFromHashes(refs: string[]): Entry<T>[];
    /**
     * Append an entry to the log.
     * @param {Entry} entry Entry to add
     * @return {Log} New Log containing the appended value
     */
    append(data: T, options?: {
        canAppend?: CanAppend<T>;
        gidSeed?: string;
        nexts?: Entry<any>[];
        pin?: boolean;
        identity?: Identity;
        signers?: ((data: Uint8Array) => Promise<SignatureWithKey>)[];
        reciever?: EncryptionTemplateMaybeEncrypted;
        onGidsShadowed?: (gids: string[]) => void;
        timestamp?: Timestamp;
    }): Promise<Entry<T>>;
    iterator(options: {
        gt?: string;
        gte?: string;
        lt?: Entry<T>[] | string;
        lte?: Entry<T>[] | string;
        amount?: number;
    }): IterableIterator<Entry<T>>;
    /**
     * Join two logs.
     *
     * Joins another log into this one.
     *
     * @param {Log} log Log to join with this Log
     * @param {number} [size=-1] Max size of the joined log
     * @returns {Promise<Log>} This Log instance
     * @example
     * await log1.join(log2)
     */
    join(log: Log<T>, options?: {
        size?: number;
        verifySignatures?: boolean;
    }): Promise<this>;
    getHeads(from: string): Entry<T>[];
    deleteRecursively(from: Entry<any>): Promise<void>;
    /**
     * Cut log to size
     * @param size
     */
    prune(size: number): void;
    removeAll(heads: Entry<any>[]): void;
    /**
     * Get the log in JSON format.
     * @returns {Object} An object with the id and heads properties
     */
    toJSON(): {
        id: string;
        heads: string[];
    };
    /**
     * Get the log in JSON format as a snapshot.
     * @returns {Object} An object with the id, heads and value properties
     */
    toSnapshot(): {
        id: string;
        heads: Entry<T>[];
        values: Entry<T>[];
    };
    /**
     * Returns the log entries as a formatted string.
     * @returns {string}
     * @example
     * two
     * └─one
     *   └─three
     */
    toString(payloadMapper?: (payload: Payload<T>) => string): string;
    /**
     * Get the log's multihash.
     * @returns {Promise<string>} Multihash of the Log as Base58 encoded string.
     */
    toMultihash(options?: {
        format?: string;
    }): Promise<string>;
    /**
     * Create a log from a hashes.
     * @param {IPFS} ipfs An IPFS instance
     * @param {Identity} identity The identity instance
     * @param {string} hash The log hash
     * @param {Object} options
     * @param {AccessController} options.access The access controller instance
     * @param {number} options.length How many items to include in the log
     * @param {Array<Entry<T>>} options.exclude Entries to not fetch (cached)
     * @param {function(hash, entry,  parent, depth)} options.onProgressCallback
     * @param {Function} options.sortFn The sort function - by default LastWriteWins
     * @returns {Promise<Log>}
     */
    static fromMultihash<T>(store: Blocks, identity: Identity, hash: string, options: {
        sortFn?: Sorting.ISortFunction;
    } & EntryFetchAllOptions<T>): Promise<Log<T>>;
    /**
     * Create a log from a single entry's hash.
     * @param {IPFS} ipfs An IPFS instance
     * @param {Identity} identity The identity instance
     * @param {string} hash The entry's hash
     * @param {Object} options
     * @param {string} options.logId The ID of the log
     * @param {AccessController} options.access The access controller instance
     * @param {number} options.length How many entries to include in the log
     * @param {Array<Entry<T>>} options.exclude Entries to not fetch (cached)
     * @param {function(hash, entry,  parent, depth)} options.onProgressCallback
     * @param {Function} options.sortFn The sort function - by default LastWriteWins
     * @return {Promise<Log>} New Log
     */
    static fromEntryHash<T>(store: Blocks, identity: Identity, hash: string | string[], options?: {
        encoding?: Encoding<T>;
        encryption?: PublicKeyEncryptionResolver;
        logId?: string;
        length?: number;
        exclude?: any[];
        shouldExclude?: (hash: string) => boolean;
        timeout?: number;
        concurrency?: number;
        sortFn?: any;
        onProgressCallback?: any;
    }): Promise<Log<T>>;
    /**
     * Create a log from a Log Snapshot JSON.
     * @param {IPFS} ipfs An IPFS instance
     * @param {Identity} identity The identity instance
     * @param {Object} json Log snapshot as JSON object
     * @param {Object} options
     * @param {AccessController} options.access The access controller instance
     * @param {number} options.length How many entries to include in the log
     * @param {function(hash, entry,  parent, depth)} [options.onProgressCallback]
     * @param {Function} options.sortFn The sort function - by default LastWriteWins
     * @return {Promise<Log>} New Log
     */
    static fromJSON<T>(store: Blocks, identity: Identity, json: {
        id: string;
        heads: string[];
    }, options?: {
        encoding?: Encoding<T>;
        encryption?: PublicKeyEncryptionResolver;
        length?: number;
        timeout?: number;
        sortFn?: Sorting.ISortFunction;
        onProgressCallback?: (entry: Entry<T>) => void;
    }): Promise<Log<T>>;
    /**
     * Create a new log from an Entry instance.
     * @param {IPFS} ipfs An IPFS instance
     * @param {Identity} identity The identity instance
     * @param {Entry|Array<Entry<T>>} sourceEntries An Entry or an array of entries to fetch a log from
     * @param {Object} options
     * @param {AccessController} options.access The access controller instance
     * @param {number} options.length How many entries to include. Default: infinite.
     * @param {Array<Entry<T>>} options.exclude Entries to not fetch (cached)
     * @param {function(hash, entry,  parent, depth)} [options.onProgressCallback]
     * @param {Function} options.sortFn The sort function - by default LastWriteWins
     * @return {Promise<Log>} New Log
     */
    static fromEntry<T>(store: Blocks, identity: Identity, sourceEntries: Entry<T>[] | Entry<T>, options: EntryFetchOptions<T> & {
        shouldExclude?: (hash: string) => boolean;
        encryption?: PublicKeyEncryptionResolver;
        sortFn?: Sorting.ISortFunction;
    }): Promise<Log<T>>;
    /**
     * Find heads from a collection of entries.
     *
     * Finds entries that are the heads of this collection,
     * ie. entries that are not referenced by other entries.
     *
     * @param {Array<Entry<T>>} entries Entries to search heads from
     * @returns {Array<Entry<T>>}
     */
    static findHeads<T>(entriesOrLogs: (Entry<T> | Log<T>)[]): Entry<T>[];
    static findTails<T>(entries: Entry<T>[]): Entry<T>[];
    static findTailHashes(entries: Entry<any>[]): string[];
    static difference<T>(a: Log<T>, b: Log<T>): Promise<Map<string, Entry<T>>>;
}
